
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdio.h>
#include <termios.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "udp_server.h"

static int udp_server_receive(void * const ref, const int fd) {
	udp_server_t * const inst = (udp_server_t * const)ref;

	uint8_t buffer[1500];

	struct sockaddr src_addr;
	socklen_t addrlen = sizeof(src_addr);

  const int len = recvfrom(inst->fd, buffer, sizeof(buffer), 0, &src_addr, &addrlen);
  if(len < 0) {
		syslog(LOG_ERR, "udp_server_receive: recvfrom: %s", strerror(errno));
		return(0);
	}

	return inst->receive_cb(src_addr, buffer, len);
}

int udp_send(const struct sockaddr * const dst_addr, const int src_port, const uint8_t * const data, const int len) {
	struct sockaddr_in addr;

  const int fd = socket(AF_INET, SOCK_DGRAM, 0);
  if(fd < 0) {
		syslog(LOG_ERR, "udp_send: socket: %s", strerror(errno));
    return(0);
	}

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(src_port);
  addr.sin_addr.s_addr = INADDR_ANY;

  if(bind(fd, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		syslog(LOG_ERR, "udp_send: bind: %s", strerror(errno));
    close(fd);
    return(0);
	}

	if(sendto(fd, data, len, 0, dst_addr, sizeof(*dst_addr)) < 0) {
		syslog(LOG_ERR, "udp_send: sendto: %s", strerror(errno));
		close(fd);
		return(0);
	}

	close(fd);
	return(1);
}

int udp_server_set_mg(udp_server_t * const inst, const char * const multiaddr, const char * const interface) {
  struct ip_mreq group;

	group.imr_multiaddr.s_addr = inet_addr(multiaddr);
	group.imr_interface.s_addr = inet_addr(interface);

	if(setsockopt(inst->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&group, sizeof(group)) < 0) {
		syslog(LOG_ERR, "udp_server_set_mg: setsockopt: %s", strerror(errno));
		close(inst->fd);
		return(0);
	}

	return(1);
}

int udp_server_open(udp_server_t * const inst, const int port, udp_receive_cb_t receive_cb, const char * const device) {
	inst->receive_cb = receive_cb;

  if((inst->fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		syslog(LOG_ERR, "udp_server_open: socket: %s", strerror(errno));
		return(0);
	}

	if(device) {
		if(setsockopt(inst->fd, SOL_SOCKET, SO_BINDTODEVICE, device, strlen(device))) {
			syslog(LOG_ERR, "udp_server_open: setsockopt bindtodevice %s: %s", device, strerror(errno));
			return(0);
		}
	}

	int optval = 1;
 	if(setsockopt(inst->fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval))) {
		syslog(LOG_ERR, "udp_server_open: setsockopt reuseaddr: %s", strerror(errno));
		return(0);
	}
   
  struct sockaddr_in host;
  host.sin_family = AF_INET;
  host.sin_addr.s_addr = INADDR_ANY;
  host.sin_port = htons(port);
  
  if(bind(inst->fd, (struct sockaddr *) &host, sizeof(host)) == -1) {
		syslog(LOG_ERR, "udp_server_open: bind: %s", strerror(errno));
		close(inst->fd);
    return(0);
  }

	if(select_register(inst->fd, udp_server_receive, NULL, NULL, inst) < 0) {
		syslog(LOG_ERR, "udp_server_open: select_register: out of memory");
		close(inst->fd);
		return(0);
	}

	return(1);
}

